{# #######################
  includes.html
##########################

This is printed in the HTML head section of the report and includes all of
the CSS and JavaScript dependencies (plus favicon images).

Note - to make the report stand along (not requiring any associated files),
it prints the contents of these files into the report.

#}

<!-- Favicon includes -->
<link rel="icon" type="image/png" sizes="32x32" href="assets/img/favicon-32x32.png">
<link rel="icon" type="image/png" sizes="96x96" href="assets/img/favicon-96x96.png">
<link rel="icon" type="image/png" sizes="16x16" href="assets/img/favicon-16x16.png">

<!-- Include CSS -->
<style type="text/css">
@font-face{
  font-family:'Glyphicons Halflings';
  src:url('assets/fonts/glyphicons-halflings-regular.eot');
  src:url('assets/fonts/glyphicons-halflings-regular.eot') format('embedded-opentype'),
      url('assets/fonts/glyphicons-halflings-regular.woff2') format('woff2'),
      url('assets/fonts/glyphicons-halflings-regular.woff') format('woff'),
      url('assets/fonts/glyphicons-halflings-regular.ttf') format('truetype'),
      url('assets/fonts/glyphicons-halflings-regular.svg') format('svg');
}
</style>
<link rel="stylesheet" type="text/css" href="assets/css/bootstrap.min.css">
<link rel="stylesheet" type="text/css" href="assets/css/default_multiqc.css">
{% for css_href in config.custom_css_files %}
<link rel="stylesheet" type="text/css" href="{{ css_href }}">
{% endfor %}
{%- for m in report.modules_output %}{% if m.css and m.css|length > 0 -%}{% for css_href in m.css.keys() %}
<link rel="stylesheet" type="text/css" href="{{ css_href }}">
{%- endfor %}{% endif %}{% endfor %}

<!-- Include javascript files -->
<script type="text/javascript" src="assets/js/packages/jquery-3.1.1.min.js"></script>
<script type="text/javascript" src="assets/js/packages/jquery-ui.min.js"></script>
<script type="text/javascript" src="assets/js/packages/bootstrap.min.js"></script>
<script type="text/javascript" src="assets/js/packages/highcharts.js"></script>
<script type="text/javascript" src="assets/js/packages/highcharts.heatmap.js"></script>
<script type="text/javascript" src="assets/js/packages/highcharts.exporting.js"></script>
<script type="text/javascript" src="assets/js/packages/highcharts.offline-exporting.js"></script>
<script type="text/javascript" src="assets/js/packages/highcharts.export-csv.js"></script>
<script type="text/javascript" src="assets/js/packages/jquery.tablesorter.min.js"></script>
<script type="text/javascript" src="assets/js/packages/clipboard.min.js"></script>
<script type="text/javascript" src="assets/js/packages/FileSaver.min.js"></script>
<script type="text/javascript" src="assets/js/packages/lz-string.min.js"></script>
<script type="text/javascript" src="assets/js/packages/jszip.min.js"></script>
<script type="text/javascript" src="assets/js/multiqc.js"></script>
<script type="text/javascript" src="assets/js/multiqc_tables.js"></script>
<script type="text/javascript" src="assets/js/multiqc_toolbox.js"></script>
<script type="text/javascript" src="assets/js/multiqc_plotting.js"></script>
<script type="text/javascript" src="assets/js/multiqc_mpl.js"></script>
{%- for m in report.modules_output %}{% if m.js and m.js|length > 0 -%}{% for js_href in m.js.keys() %}
<script type="text/javascript" src="{{ js_href }}"></script>
{%- endfor %}{% endif %}{% endfor %}
